@@ -75,6 +75,18 @@ sub as_txt
   ' ' . $n . $self->{style} . '> ';
   }
 
+sub _formatted_label
+  {
+  my $self = shift;
+
+  my $name = $self->label();
+  $name =~ s/\\n/\n/g;                  # insert newlines
+
+  # split into lines, but don't extranous spacing
+  my @lines = split /\n/, $name;
+  @lines;
+  }
+
 #############################################################################
 # accessor methods
 
@@ -257,11 +269,8 @@ L<Graph::Simple>.
 
 =head1 AUTHOR
 
-Copyright (C) 2004 - 2005 by Tels L<http://bloodgate.com>
+Copyright (C) 2004 - 2005 by Tels L<http://bloodgate.com>.
 
-=head1 LICENSE
-
-This library is free software; you can redistribute it and/or modify
-it under the terms of the GPL. See the LICENSE file for more details.
+See the LICENSE file for more details.
 
 =cut
